from .utils.models.segmentation_model import SegmentationModel
from .utils.train import Train
from .utils.data.dataloader import SegmentationDataLoader
from .utils.data.augmentation import MultipleColorJitter


# todo:apply augmentation


class PageExtractor(Train):
    r"""This class used to extract pages (removing borders and blank spaces around pages) from originals.

    Example

    .. code-block:: python

        >>> from batukh.tensorflow.segmenter import PageExtraction
        >>> page_Extractor = PageExtraction()
        >>> page_Extractor.load_data(train_path = "/train_data/")
        >>> page_Extractor.train(n_epochs=10,batch_size=1,weights=[1,100])
        Initializing from scratch
        Epoch: 1. Traininig: 100%|██████████| 70/70 [00:02<00:00, 23.95it/s, loss=0.0708]
        Model saved to /tf_ckpts/Fri Oct 16 08:23:13 2020/ckpt-14280
        >>> page_Extractor.save_model("/model/")
        Model saved at /saved_models


        """

    def __init__(self):
        super().__init__(model=SegmentationModel(2))
        self.train_dl = None
        self.val_dl = None

    def load_data(self, train_path, val_path=None):
        r"""Loads Train and Validation datset.

        Note:
            Respective images and labels should be of same size with same filename.
            label images should be of black background with  pixels corresponding to page area colored red.

        Args:
            train_path (str)        : path of the folder contaings originals folder (containing original images) and labels folder (containing label images) for train dataset.
            val_path (str,optional) : path of the folder contaings originals folder (containing original images) and labels folder (containing label images) for validation dataset.
            """
        self.train_dl = SegmentationDataLoader(
            train_path, self.model.n_classes)
        if val_path is not None:
            self.val_dl = SegmentationDataLoader(
                val_path, self.model.n_classes)

    def train(self, n_epochs, train_dl=None, val_dl=None, batch_size=1, repeat=1, criterion=None, class_weights=None, optimizer=None, weight_decay=None, learning_rate=0.0001, lr_decay=None, save_checkpoints=True, checkpoint_freq=None, checkpoint_path=None, max_to_keep=5, log_freq=100):
        if train_dl is None:
            train_dl = self.train_dl
        if val_dl is None:
            val_dl = self.val_dl
        super().train(n_epochs, train_dl=train_dl, val_dl=val_dl, batch_size=batch_size, repeat=repeat, criterion=criterion, class_weights=class_weights,
                      optimizer=optimizer, weight_decay=weight_decay, learning_rate=learning_rate, lr_decay=lr_decay, save_checkpoints=save_checkpoints, checkpoint_freq=checkpoint_freq, checkpoint_path=checkpoint_path, max_to_keep=max_to_keep, log_freq=log_freq)


class ImageExtractor(Train):
    r"""The class used to extract images.

    Example

    .. code-block:: python

        >>> from batukh.tensorflow.segmenter import ImageExtractor
        >>> image_Extractor = ImageExtractor()
        >>> image_Extractor.load_data( train_path="/train_data/",val_path="/val_data/")
        >>> image_Extractor.train(n_epochs=1)
        Initializing from scratch
        Epoch: 1. Traininig: 100%|██████████| 70/70 [00:02<00:00, 23.95it/s, loss=0.0708]
        Epoch: 1. validation: 100%|██████████| 70/70 [00:02<00:00, 23.95it/s, loss=0.0708]
        Model saved to /tf_ckpts/Fri Oct 16 08:23:13 2020/ckpt-14280

        """

    def __init__(self):
        super().__init__(model=SegmentationModel(2))
        self.train_dl = None
        self.val_dl = None

    def load_data(self, train_path, val_path=None):
        r"""Loads Train and Validation datset.

        Note:
            Respective images and labels should be of same size with same filename.
            label images should be of black background with  pixels corresponding image areas colored red.

        Args:
            train_path (str)        : path of the folder contaings originals folder (containing originals images) and labels folder (containing label images) for train dataset.
            val_path (str,optional) : path of the folder contaings originals folder (containing originals images) and labels folder (containing label images) for validation dataset.
            """
        self.train_dl = SegmentationDataLoader(
            train_path, self.model.n_classes)
        if val_path is not None:
            self.val_dl = SegmentationDataLoader(
                val_path, self.model.n_classes)

    def train(self, n_epochs, train_dl=None, val_dl=None, batch_size=1, repeat=1, criterion=None, class_weights=None, optimizer=None, weight_decay=None, learning_rate=0.0001, lr_decay=None, save_checkpoints=True, checkpoint_freq=None, checkpoint_path=None, max_to_keep=5, log_freq=100):
        if train_dl is None:
            train_dl = self.train_dl
        if val_dl is None:
            val_dl = self.val_dl

        super().train(n_epochs, train_dl=train_dl, val_dl=val_dl, batch_size=batch_size, repeat=repeat, criterion=criterion, class_weights=class_weights,
                      optimizer=optimizer, weight_decay=weight_decay, learning_rate=learning_rate, lr_decay=lr_decay, save_checkpoints=save_checkpoints, checkpoint_freq=checkpoint_freq, checkpoint_path=checkpoint_path, max_to_keep=max_to_keep, log_freq=log_freq)


class LayoutExtractor(Train):
    r"""This class is used to extract diffrent layouts from a image.

    Example

    .. code-block:: python

        >>> from batukh.tensorflow.segmenter import LayoutExtractor
        >>> layout_Extractor = LayoutExtractor(2)
        >>> layout_Extractor.load_data(train_path ="/train_data/",val_data="/val_data/")
        >>> layout_Extractor.train(n_epochs=1,checkpoint_path="/tf_chkpts/")
        Restored from /tf_chkpts/ckpt-13280
        Epoch: 1. Traininig: 100%|██████████| 70/70 [00:02<00:00, 23.95it/s, loss=0.0708]
        Epoch: 1. validation: 100%|██████████| 70/70 [00:02<00:00, 23.95it/s, loss=0.0708]
        Model saved to /tf_ckpts/Fri Oct 16 08:23:13 2020/ckpt-14280

        """

    def __init__(self, n_layouts):
        super().__init__(model=SegmentationModel(n_layouts))
        self.train_dl = None
        self.val_dl = None

    def load_data(self, train_path, val_path=None):
        r"""Loads Train and Validation datset.

        Note:
            Respective images and labels should bepython of same size with same filename.
            label images should be of black background with  pixels corresponding to diffrent areas colored diffrently.

        Args:
            train_path (str)        : path of the folder contaings originals folder (containing originals images) and labels folder (containing label images) for train dataset.
            val_path (str,optional) : path of the folder contaings originals folder (containing originals images) and labels folder (containing label images) for validation dataset.
            """
        self.train_dl = SegmentationDataLoader(
            train_path, self.model.n_classes)
        if val_path is not None:
            self.val_dl = SegmentationDataLoader(
                val_path, self.model.n_classes)

    def train(self, n_epochs, train_dl=None, val_dl=None, batch_size=1, repeat=1, criterion=None, class_weights=None, optimizer=None, weight_decay=None, learning_rate=0.0001, lr_decay=None, save_checkpoints=True, checkpoint_freq=None, checkpoint_path=None, max_to_keep=5, log_freq=None):
        if train_dl is None:
            train_dl = self.train_dl
        if val_dl is None:
            val_dl = self.val_dl
        super().train(n_epochs, train_dl=train_dl, val_dl=val_dl, batch_size=batch_size, repeat=repeat, criterion=criterion, class_weights=class_weights,
                      optimizer=optimizer, weight_decay=weight_decay, learning_rate=learning_rate, lr_decay=lr_decay, save_checkpoints=save_checkpoints, checkpoint_freq=checkpoint_freq, checkpoint_path=checkpoint_path, max_to_keep=max_to_keep, log_freq=log_freq)


class BaselineDetector(Train):
    r"""This class is used to detect baseline.

    Example

    .. code-block:: python

        >>> from batukh.tensorflow.segmenter import BaselineDetector
        >>> baseline_Detector = BaselineDetector()
        >>> baseline_Detector.load_data("/train_data/")
        >>> baseline_Detector.train(1,weights=[1:700])
        Initializing from scratch
        Epoch: 1. Traininig: 100%|██████████| 70/70 [00:02<00:00, 23.95it/s, loss=0.0708]
        Model saved to /tf_ckpts/Fri Oct 16 08:23:13 2020/ckpt-14280

        """

    def __init__(self):
        super().__init__(model=SegmentationModel(2))
        self.train_dl = None
        self.val_dl = None

    def load_data(self, train_path, val_path=None):
        r"""Loads Train and Validation datset.

        Note:
            Respective images and labels should be of same size with same filename.
            label images should be of black background with  red lines of about 5px representing baselines.

        Args:
            train_path (str)        : path of the folder contaings originals folder (containing original images) and labels folder (containing label images) for train dataset.
            val_path (str,optional) : path of the folder contaings originals folder (containing original images) and labels folder (containing label images) for validation dataset.
            """
        self.train_dl = SegmentationDataLoader(
            train_path, self.model.n_classest)
        if val_path is not None:
            self.val_dl = SegmentationDataLoader(
                val_path, self.model.n_classes)

    def train(self, n_epochs, train_dl=None, val_dl=None, batch_size=1, repeat=1, criterion=None, class_weights=None, optimizer=None, weight_decay=None, learning_rate=0.0001, lr_decay=None, save_checkpoints=True, checkpoint_freq=None, checkpoint_path=None, max_to_keep=5, log_freq=100):
        if train_dl is None:
            train_dl = self.train_dl
        if val_dl is None:
            val_dl = self.val_dl
        super().train(n_epochs, train_dl=train_dl, val_dl=val_dl, batch_size=batch_size, repeat=repeat, criterion=criterion, class_weights=class_weights,
                      optimizer=optimizer, weight_decay=weight_decay, learning_rate=learning_rate, lr_decay=lr_decay, save_checkpoints=save_checkpoints, checkpoint_freq=checkpoint_freq, checkpoint_path=checkpoint_path, max_to_keep=max_to_keep, log_freq=log_freq)



import torch
from torch import nn, optim
from torchvision import transforms
import os
import random

from .utils.data.dataloader import OCRDataLoader

from .utils.models.ocr_model import ImgEncoder, Encoder, AttnDecoderRNN

from os.path import join
from tqdm import tqdm
from time import localtime

MAX_LENGTH = 50


class OCR:
    r"""
    A basic CRNN which detects the characters in an image.

    Note:
        The :attr:`encoder_input` has been set to 1536 as default. This has
        been done by keeping in mind that the height of each image is 60px. If
        your images have a different height, say, :math:`h`, then encoder_input
        should be 
        :math:`\left \lfloor \frac{\left \lfloor \frac{h-4}{2} \right \rfloor -4}{2} \right \rfloor \times 128`.

    Args:
        encoder_input (int, optional): Input size of vector that goes into the encoder.
            Default: 1536, based on the assumption that the input shape is
            `[1, 3, 60, x]`.
        encoder_hidden (int, optional): hidden size of the GRU layer of encoder. 
            Default: 128.
        encoder_nlayers (int, optional): number of layers in the GRU layer of encoder.
            Default: 2.
        decoder_hidden (int, optional): hidden size of the GRU layer of decoder.
            Default: 128.
        decoder_nlayers (int, optional): number of layers in the GRU layer of decoder.
            Default: 2.
        decoder_output (int, optional): Size of the output vector
            (i.e the number of total characters in your language, including EOS and SOS).
            Default: ``None``.
        dropout (float, optional): The probability of the dropout layers.
            Default: 0.1.
        max_length (int, optional): The maximum number of characters that can occur in an image.
            Default: 50.
    """
# TODO: Replace ``encoder_input`` with ``image_size``.
# TODO: Have an EOS and SOS option available during training.
# maybe the custom dataloader does not have an EOS or SOS varaible.

    def __init__(self,
                 encoder_input=1536,
                 encoder_hidden=128,
                 encoder_nlayers=2,
                 decoder_hidden=128,
                 decoder_nlayers=2,
                 decoder_output=None,
                 dropout=0.1,
                 max_length=MAX_LENGTH,
                 ):

        # if providing decoder_output as None, they need to use load_data
        # Only provide decoder_output, if using custom data loaders.

        self.max_length = max_length

        self.img_encoder = ImgEncoder()

        self.encoder = Encoder(encoder_input,
                               encoder_hidden,
                               encoder_nlayers)
        self.decoder = AttnDecoderRNN(decoder_hidden,
                                      decoder_output,
                                      decoder_nlayers,
                                      dropout,
                                      max_length)

    def load_data(self,
                  train_dir,
                  train_label_path,
                  val_dir=None,
                  val_label_path=None,
                  transform=None):
        """
        Args:
            train_dir (str): directory containing training images.
            train_label_path (str): path to a file with training labels.
            val_dir (str, optional): directory containing validation images.
                Default: None
            val_label_path (str, optional): path to a file with training labels.
                Default: None
            transform (:class:`~torchvision.transforms`, optional): transforms to be applied on images.
                Default: None       
        """

        self.train_dl = OCRDataLoader(train_dir, train_label_path, transform)

        # if loading data, then make decoder_output
        decoder_output = len(self.train_dl.letter2index)
        self.decoder.embedding = nn.Embedding(
            decoder_output, self.decoder.hidden_size)
        self.decoder.out = nn.Linear(self.decoder.hidden_size, decoder_output)
        self.decoder.output_size = decoder_output

        if val_dir is None or val_label_path is None:
            self.val_dl = None
        else:
            self.val_dl = OCRDataLoader(val_dir, val_label_path, transform)

    def forward_step(self,
                     x,
                     y,
                     criterion,
                     device=None,
                     teacher_forcing_ratio=0.1):
        if device is None:
            device = torch.device(
                "cuda" if torch.cuda.is_available() else "cpu")

        loss = 0

        out = self.img_encoder(x)
        # conv_output shape will be [1,1536, x']

        enc_hidden = self.encoder.initHidden()

        # for attention
        encoder_outputs = torch.zeros(
            self.max_length, self.encoder.hidden_size, device=device)

        # pass each of the [1, 1536] vectors to encoder
        for idx in range(out.shape[2]):
            enc_out, enc_hidden = self.encoder(out[:, :, idx], enc_hidden)
            encoder_outputs[idx] = enc_out[0, 0]

        decoder_input = torch.tensor(
            [[[self.train_dl.SOS]]], device=device)

        dec_hidden = enc_hidden

        use_teacher_forcing = random.random() < teacher_forcing_ratio

        target_len = len(y)

        if use_teacher_forcing:
            for di in range(target_len):
                dec_out, dec_hidden, dec_attn = self.decoder(
                    decoder_input, dec_hidden, encoder_outputs)
                loss += criterion(dec_out, y[di])
                decoder_input = y[di]  # teacher forcing
        else:
            for di in range(target_len):
                dec_out, dec_hidden, dec_attn = self.decoder(
                    decoder_input, dec_hidden, encoder_outputs)

                topv, topi = dec_out.topk(1)
                decoder_input = topi.squeeze().detach()

                loss += criterion(dec_out, y[di])

                if decoder_input.item() == self.train_dl.EOS:
                    break

        return loss, target_len

# TODO: before creating optimizers make sure they do not already exist
    def initOptimizers(self,
                       imgenc_optimizer=None,
                       enc_optimizer=None,
                       dec_optimizer=None,
                       learning_rate=0.001,):

        if imgenc_optimizer is None:
            imgenc_optimizer = optim.Adam(
                self.img_encoder.parameters(), lr=learning_rate)
        if enc_optimizer is None:
            enc_optimizer = optim.Adam(
                self.encoder.parameters(), lr=learning_rate)
        if dec_optimizer is None:
            dec_optimizer = optim.Adam(
                self.decoder.parameters(), lr=learning_rate)
        return imgenc_optimizer, enc_optimizer, dec_optimizer

    def train(self,
              n_epochs,
              train_dl=None,
              val_dl=None,
              shuffle=True,
              criterion=None,
              imgenc_optimizer=None,
              enc_optimizer=None,
              dec_optimizer=None,
              learning_rate=0.001,
              save_checkpoints=True,
              checkpoint_freq=None,
              checkpoint_path="./",  # Change to None
              max_to_keep=5,
              device=None,
              ):
        r"""
        Args:
            n_epochs (int): number of epochs to train for.
            train_dl (:class:`~torch.utils.data.DataLoader`, optional): dataloader
                to use for training. Default None.
            val_dl (:class:`~torch.utils.data.DataLoader`, optional): dataloader
                to use for validation. Default None.
            shuffle (bool, optional): whether to shuffle the default dataloader.
                Default: True.
            criterion (:class:`~torch.nn.module.loss._Loss` or :class:`~torch.nn.module.loss._WeightedLoss`, optional): 
                The loss function to use.
                Default: ``CrossEntropyLoss(reduction="mean")``
            imgenc_optimizer (:class:`~torch.optim.Optimizer`, optional): 
                optimizer for img_encoder.
                Default: Adam.
            enc_optimizer (:class:`~torch.optim.Optimizer`, optional): 
                optimizer for encoder.
                Default: Adam.
            dec_optimizer (:class:`~torch.optim.Optimizer`, optional): 
                optimizer for decoder.
                Default: Adam.
            learning_rate (int, optional): learning rate to be used for default
                optimizers.
                Default: 0.001.
            save_checkpoints (bool, optional): whether to save checkpoints or
                not.
                Default: True.
            checkpoint_freq (int, optional): the frequency of saving saving
                checkpoints.
                Default: :math:`\left \lfloor\frac{\text{n_epochs}}{10}\right \rfloor + 1`.
            checkpoint_path (str, optional): Path of the checkpoint folder.
                Default: "./"
            max_to_keep (int or None, optional): The maximum number of latest 
                checkpoints to be saved. ``None`` to save all. 
                Default: 5.
            device (str, optional): The device to do computations on.
                Default: GPU, if GPU is available, else CPU. 
        """

        if device is None:
            device = torch.device(
                "cuda" if torch.cuda.is_available() else "cpu")

        self.img_encoder.to(device)
        self.encoder.to(device)
        self.decoder.to(device)

        imgenc_optimizer, enc_optimizer, dec_optimizer = self.initOptimizers(
            imgenc_optimizer, enc_optimizer, dec_optimizer, learning_rate
        )

        if criterion is None:
            criterion = nn.CrossEntropyLoss()

        checkpoint_path = join(checkpoint_path, "checkpoints")
        os.makedirs(checkpoint_path, exist_ok=True)
        if checkpoint_freq is None:
            checkpoint_freq = n_epochs//10+1

        for epoch in range(n_epochs):
            if train_dl is None:
                if getattr(self, "train_dl", None) is None:
                    raise Exception(
                        "No DataLoader found. Either pass one in train or use load_data method.")
                else:
                    train_dl = self.train_dl(1, shuffle)
            if val_dl is None:
                if getattr(self, "val_dl", None) is None:
                    val_dl = None
                else:
                    val_dl = self.val_dl(1, shuffle)

            self.img_encoder.train()
            self.encoder.train()
            self.decoder.train()
            total_loss = 0

            # Progress bar
            pbar = tqdm(total=len(train_dl))
            pbar.set_description(f"Epoch: {epoch}. Traininig")

            for i, (x, y) in enumerate(train_dl):

                if x.shape[3] < 16:
                    pbar.update(1)
                    continue

                imgenc_optimizer.zero_grad()
                enc_optimizer.zero_grad()
                dec_optimizer.zero_grad()
                # maybe add teacher_force_ratio control
                x = x.to(device)
                y = y.to(device)
                y = y[0]

                loss, target_len = self.forward_step(
                    x, y, criterion, device)

                loss.backward()

                imgenc_optimizer.step()
                enc_optimizer.step()
                dec_optimizer.step()

                total_loss = total_loss + (loss.item()/target_len)

                pbar.update()
                pbar.set_postfix(loss=total_loss/(i+1))
            pbar.close()

            if val_dl is not None:
                self.img_encoder.eval()
                self.encoder.eval()
                self.decoder.eval()
                eval_loss = 0

                # validation progress bar
                pbar = tqdm(total=len(val_dl))
                pbar.set_description(f"Epoch: {epoch}. Validating")

                for i, (x, y) in enumerate(val_dl):
                    if x.shape[3] < 16:
                        pbar.update(1)
                        continue

                    x = x.to(device)
                    y = y.to(device)
                    y = y[0]  # only

                    loss, target_len = self.forward_step(
                        x, y, criterion, device)

                    eval_loss = eval_loss + (loss.item()/target_len)

                    pbar.update(1)
                    pbar.set_postfix(loss=eval_loss/(i+1))
                pbar.close()
            if epoch % checkpoint_freq == 0:
                self.save_model(checkpoint_path, epoch)

    def predict(self, x, device=None):
        r"""
        predicts the output given a tensor.

        Args:
            x (:class:`~torch.Tensor`): tensor of shape ``[1, 3, width, height]``.
            device (str, optional): device to predict on.
                Default: GPU if it is available else CPU.

        Returns:
            list: sequence of outputs.
        """
        self.img_encoder.eval()
        self.encoder.eval()
        self.decoder.eval()

        if device is None:
            device = torch.device(
                "cuda" if torch.cuda.is_available() else "cpu")

        outputs = []

        out = self.img_encoder(x)

        enc_hidden = self.encoder.initHidden()

        encoder_outputs = torch.zeros(
            self.max_length, self.encoder.hidden_size, device=device)

        for idx in range(out.shape[2]):
            enc_out, enc_hidden = self.encoder(out[:, :, idx], enc_hidden)
            encoder_outputs[idx] = enc_out[0, 0]

        decoder_input = torch.tensor(
            [[[self.train_dl.SOS]]], device=device)

        dec_hidden = enc_hidden

        while True:
            dec_out, dec_hidden, dec_attn = self.decoder(
                decoder_input, dec_hidden, encoder_outputs)

            topv, topi = dec_out.topk(1)
            decoder_input = topi.squeeze().detach()

            outputs.append(decoder_input)
            if decoder_input.item() == self.train_dl.EOS:
                break

        return outputs

    def load_model(self, path):
        models = torch.load(path)
        print(self.img_encoder.load_state_dict(models["img_encoder"]))
        print(self.encoder.load_state_dict(models["encoder"]))
        print(self.decoder.load_state_dict(models["decoder"]))
        print("Models Loaded!")

    def save_model(self, path, postfix=0):
        name = "{} {}-{}-{} {}.{}.{}.pt".format(postfix, *localtime()[:6])
        models = {"img_encoder": self.img_encoder.state_dict(),
                  "encoder": self.encoder.state_dict(),
                  "decoder": self.decoder.state_dict()}
        torch.save(models, join(path, name))
        print("Models Saved!")


# TODO: Test each and every module of this file.



from .utils.train import Train
from .utils.data.dataloader import OCRDataLoader
from .utils.data.augmentation import MultipleColorJitter
from .utils.models.ocr_model import OCRModel
import tensorflow as tf

# todo batch size segmenter.
# todo dynamic  weights.
# todo checkpoints.
# lr sheduler.
# l2 regulirization e-6.

# todo tensorboard
# todo example boun
#


class OCR(Train):
    r"""This class is used for octical character recognition.

    Example

    .. code:: python

        >>> from batukh.tensorflow.ocr import OCR
        >>> m = OCR(177)
        >>> m.load_data(train_path="/data/",height=32)
        >>> m.train(1)


    Args:
        n_classes (int) : number of outputs nodes of ocr model.
    """

    def __init__(self, n_classes):

        super().__init__(model=OCRModel(n_classes), is_ocr=True)
        self.train_dl = None
        self.val_dl = None

    def load_data(self, train_path, val_path=None, height=64):
        r"""Loads Train and Validation datset.

        Args:
            train_path (str)        : path of the folder contaings images folder,labels.txt and table.txt for train dataset.
            val_path (str,optional) : path of the folder contaings images folder ,labels.txt and table.txt  for validation dataset.
            """
        self.train_dl = OCRDataLoader(
            train_path, height)
        if val_path is not None:
            self.val_dl = OCRDataLoader(
                val_path, height)

    def train(self, n_epochs, train_dl=None, val_dl=None, repeat=1, criterion=None, class_weights=None, optimizer=None, learning_rate=0.0001, save_checkpoints=True, checkpoint_freq=None, checkpoint_path=None, max_to_keep=5):
        if train_dl is None:
            train_dl = self.train_dl
        if val_dl is None:
            val_dl = self.val_dl
        super().train(n_epochs, train_dl=train_dl, val_dl=val_dl, batch_size=1, repeat=repeat, criterion=criterion, class_weights=class_weights,
                      optimizer=optimizer, learning_rate=learning_rate, save_checkpoints=save_checkpoints, checkpoint_freq=checkpoint_freq, checkpoint_path=checkpoint_path, max_to_keep=max_to_keep)

    def map2string(self, inputs, table=None, blank_index=None):
        # todo : shift theese methods to ocr
        """Maps tensor to stings as per :class:`~self.inv_table`.

        Args:
            inputs (:class:`tensorflow.SparseTensor`) : Input tensors.
            table ( :class:`tensorflow.lookup.StaticHashTable`,optional) : Table according to which maping is done.
                Default: ``self.train_dl.inv_table``
            blank_index (int,optional) : Blank index.
                Default : ``self.train_dl.blank_index``



        Returns:
            list : list of strings."""
        if table is None:
            table = self.train_dl.inv_table
        if blank_index is None:
            blank_index = self.train_dl.blank_index
        inputs = tf.sparse.to_dense(inputs,
                                    default_value=blank_index).numpy()
        strings = []
        for i in inputs:
            text = [table[char_index] for char_index in i
                    if char_index != blank_index]
            strings.append(''.join(text))
        return strings

    def decode(self, inputs, from_pred=True, method='gready', merge_repeated=True, table=None, blank_index=None):
        """Decodes the model logits using ctc decoder.

        Example

        .. code:: python

            >>> from batukh.tensorflow.ocr import OCR
            >>> import tensorflow as tf
            >>> m = OCR(177)
            >>> m.load_model("/saved_model/")
            >>> x = tf.io.read_file("/image.png")
            >>> x = tf.io.decode_png(x,channels=1)
            >>> y = m.predict(x)
            >>> pred = m.decode(y)



        Args:
            inputs ( :class:`tensorflow.Tensor`) : Input tensor.
            from_pred (bol,optional): ``True`` if input is return of ( :class:`~batukh.tensorflow.ocr.OCR.predict`. ``False`` if input is a :class:`tensorflow.SparseTensor`.
                Default: ``True``
            method (str,optional)   :if  ``'gready'``  :class:`tensorflow.nn.ctc_greedy_decoder` used for decoding.if  ``'beam_search'``  :class:`tensorflow.nn.ctc_beam_search_decoder` used.
                Default: `` 'greedy' ``
            merge_repeated (bol,optional): Specifes if similar charsters will be merged.
                Default: ``True``
            table ( :class:`tensorflow.lookup.StaticHashTable`,optional) : Table according to which maping is done.
                Default: ``self.train_dl.inv_table``
            blank_index (int,optional) : Blank index.
                Default : ``self.train_dl.blank_index``

        Returns:
            list: decoded list of strings  """
        self.merge_repeated = merge_repeated
        if from_pred:
            logit_length = tf.fill([tf.shape(inputs)[0]], tf.shape(inputs)[1])
            if method == 'greedy':
                decoded, _ = tf.nn.ctc_greedy_decoder(
                    inputs=tf.transpose(inputs, perm=[1, 0, 2]),
                    sequence_length=logit_length,
                    merge_repeated=self.merge_repeated)
            elif method == 'beam_search':
                decoded, _ = tf.nn.ctc_beam_search_decoder(
                    inputs=tf.transpose(inputs, perm=[1, 0, 2]),
                    sequence_length=logit_length)
            inputs = decoded[0]

        decoded = self.map2string(decoded)
        return decoded



import torch
from torch import nn
import torch.nn.functional as F


# The model architechture is organized into three modules:
# 1. `ImgEncoder`: A convolutional network.
# -  It will take input of an image of `[3, 60, x]` (channels, height, width)
# and produce the output of `[128, 12, x']`, which is then flattened out
# to `[1536, x']`. Minimum height of image, x = 16.

# 2. `Encoder(input_size, hidden_size, n_layers)`:
# - It has a linear layer followed by a GRU layer.
# - It takes a vector of `[1, input_size]` and the linear converts into
# `[1, hidden_size]` vector.
# - Then the `[1, hidden_size]` vector is passed via a GRU with `n_layers`.
# - The hidden input to GRU is initialized as zeros.

# 3. `AttnDecoderRNN(hidden_size, output_size, n_layers)`:
# - It has an embedding layer giving out `hidden_size` length of vectors.
# - Then a GRU layer with `hidden_size` output and `n_layers`.
# - Then a linear layer of `output_size` nodes.
# - Then a softmax.


class ImgEncoder(nn.Module):
    def __init__(self):
        super(ImgEncoder, self).__init__()
        self.conv1 = nn.Conv2d(3, 16, kernel_size=3)
        self.conv2 = nn.Conv2d(16, 32, kernel_size=3)
        self.conv3 = nn.Conv2d(32, 64, kernel_size=3)
        self.conv4 = nn.Conv2d(64, 128, kernel_size=3)

        self.pool = nn.MaxPool2d((2, 2))
        self.flatten = nn.Flatten(end_dim=-2)

    def forward(self, x):

        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        x = self.pool(x)

        x = F.relu(self.conv3(x))
        x = F.relu(self.conv4(x))

        x = self.pool(x)
        x = self.flatten(x)
        return x


class Encoder(nn.Module):
    def __init__(self, input_size, hidden_size, n_layers, device=None):

        super(Encoder, self).__init__()

        if device is None:
            device = torch.device(
                "cuda" if torch.cuda.is_available() else "cpu")
        self.device = device
        self.n_layers = n_layers

        self.hidden_size = hidden_size

        self.fc1 = nn.Linear(input_size, hidden_size)

        # self.em = nn.Embedding(input_size, hidden_size)
        self.gru = nn.GRU(hidden_size, hidden_size, num_layers=n_layers)

    def forward(self, input, hidden):

        output = F.relu(self.fc1(input)).view(1, 1, -1)

        # embedded = self.em(input).view(1, 1, -1)
        # output = embedded
        output, hidden = self.gru(output, hidden)

        return output, hidden

    def initHidden(self):
        return torch.zeros(self.n_layers, 1, self.hidden_size,
                           device=self.device)


class AttnDecoderRNN(nn.Module):

    # TODO: Check if we can dynamically create a linear layer in forward method
    # to replace `self.attn` and eliminate the need for `MAX_LENGTH`
    def __init__(self, hidden_size, output_size, n_layers, dropout_p,
                 max_length, device=None):
        super(AttnDecoderRNN, self).__init__()

        if device is None:
            device = torch.device(
                "cuda" if torch.cuda.is_available() else "cpu")
        self.device = device

        self.hidden_size = hidden_size
        self.output_size = output_size
        self.dropout_p = dropout_p
        self.max_length = max_length
        self.n_layers = n_layers

        if self.output_size is not None:
            self.embedding = nn.Embedding(self.output_size, self.hidden_size)
        self.attn = nn.Linear(self.hidden_size * 2, self.max_length)
        self.attn_combine = nn.Linear(self.hidden_size * 2, self.hidden_size)
        self.dropout = nn.Dropout(self.dropout_p)
        self.gru = nn.GRU(self.hidden_size, self.hidden_size, self.n_layers)
        if self.output_size is not None:
            self.out = nn.Linear(self.hidden_size, self.output_size)

    def forward(self, input, hidden, encoder_outputs):
        embedded = self.embedding(input).view(1, 1, -1)
        embedded = self.dropout(embedded)

        attn_weights = F.softmax(
            self.attn(torch.cat((embedded[0], hidden[0]), 1)), dim=1)
        attn_applied = torch.bmm(attn_weights.unsqueeze(0),
                                 encoder_outputs.unsqueeze(0))

        output = torch.cat((embedded[0], attn_applied[0]), 1)
        output = self.attn_combine(output).unsqueeze(0)

        output = F.relu(output)
        output, hidden = self.gru(output, hidden)

        output = self.out(output[0])
        return output, hidden, attn_weights

    def initHidden(self):
        return torch.zeros(self.n_layers, 1, self.hidden_size, device=self.device)



from torch.utils.data import Dataset, DataLoader
import os
from os.path import join
import torch
from PIL import Image
from torchvision import transforms


class SegmentationDataLoader(Dataset):
    """
    Makes a :class:`~torch.utils.data.Dataset` and a :class:`~torch.utils.data.DataLoader` 
    on call which can be iterated with
    a given batch size.

    Args:
        input_dir (str): path to the directory containing the input images.
        label_dir (str): path to the directory containing the labelled images.
        transforms (transforms or None, optional): transforms that act on a 
            tuple of images and return a tuple of transformed images.
            For some examples, see :class:`~batukh.torch_implem.data.augmentation.MultipleRandomRotation`  
    """

    def __init__(self, input_dir, label_dir, transforms=None):

        self.transforms = transforms
        self.input_dir = input_dir
        self.label_dir = label_dir

        self.input_files = sorted(os.listdir(input_dir))
        self.label_files = sorted(os.listdir(label_dir))

        assert self.input_files == self.label_files

    def __len__(self):
        return len(self.input_files)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        input_image = Image.open(join(
            self.input_dir, self.input_files[idx]))
        # input_image = input_image.resize((1024, 1536))
        label_image = Image.open(join(
            self.label_dir, self.label_files[idx]))
        # label_image = label_image.resize((1024, 1536))

        if self.transforms is not None:
            input_image, label_image = self.transforms(
                (input_image, label_image))

        label_image = (label_image[0, :, :] > (50/255))*1

        return input_image, label_image

    def __call__(self, batch_size=1, shuffle=True, num_workers=4, pin_memory=True):
        """
        Args:
            batch_size (int, optional): size of the each batch for every iteration.
                Default: 1
            shuffle (bool, optional): Whether to shuffle the dataset or not. Default ``True``.

        Returns:
            :class:`~torch.utils.data.DataLoader` : Instance
        """
        return DataLoader(self, batch_size=batch_size, shuffle=shuffle,
                          num_workers=num_workers, pin_memory=pin_memory)


class OCRDataLoader(Dataset):
    r"""
    Args:
        image_dir (str): Path to the directory containing images with names like `1.png, 2.png`.
        labels_path (str): Path to a file containing labels like "`1. label_1\\n2. label_2`".
        transform (:class:`torchvision.transforms`, optional): A transform 
           (or a composition of multiple transforms) to be applied on the input images.
    """

    def __init__(self, image_dir, labels_path, transform=None):

        if transform is None:
            transform = transforms.Compose([transforms.Grayscale(3),
                                            transforms.ToTensor()])
        self.transform = transform
        self.image_dir = image_dir
        self.files = os.listdir(image_dir)
        self.files.sort(key=lambda x: int(x.split(".")[0]))

        with open(labels_path) as f:
            self.labels = f.readlines()

        # add assertion for checking filenames and label numbers
        assert len(self.files) == len(self.labels)

        print("Building Dictionary. . .")
        self.SOS = 0
        self.EOS = 1

        self.labels = list(
            map(lambda x: x.split(":", 1)[-1].strip(), self.labels))

        self.index2letter = dict(enumerate(set("".join(self.labels)), 2))
        self.index2letter[self.SOS] = "<SOS>"
        self.index2letter[self.EOS] = "<EOS>"
        self.letter2index = {v: k for k, v in self.index2letter.items()}

    def __call__(self, batch_size=1, shuffle=True):
        return DataLoader(self, batch_size=batch_size, shuffle=shuffle)

    def __len__(self):
        return len(self.files)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        image = Image.open(os.path.join(self.image_dir, self.files[idx]))
        label = self.labels[idx]

        image = self.transform(image)
        label = self.transform_label(label)

        return image, label

    def transform_label(self, label):
        """
        Encodes a label into its vector given the dictionary. Also adds the
        `End of Sentence` value to it.

        Args:
            label (str): A string to be encoded to a vector.

        Returns:
            :class:`~torch.tensor`: Encoded vector.
        """
        label = [self.letter2index[letter] for letter in label]
        label.append(self.EOS)
        return torch.tensor(label, dtype=torch.long).view(-1, 1)



import tensorflow as tf


class MultipleColorJitter():
    def __init__(self, ds, brightness=0, contrast=0, saturation=0, hue=0, n_max=None, prob=0.2):
        self.dataset = ds
        self.prob = prob
        self.n_max = n_max
        self.brightness = brightness
        self.contrast = contrast
        self.saturation = saturation
        self.hue = hue

    def jitter(self, batch):
        if tf.random.uniform([], 0, 1) <= self.prob:
            for i in range(self.n_max):

                batch[i] = tf.image.adjust_brightness(
                    batch[i], tf.random.uniform([], -1*self.brightness, self.brightness))
                batch[i] = tf.image.adjust_contrast(
                    batch[i], tf.random.uniform([], -1*self.contrast, self.contrast))
                batch[i] = tf.image.adjust_hue(
                    batch[i], tf.random.uniform([], -1*self.hue, self.hue))
                batch[i] = tf.image.adjust_saturation(
                    batch[i], tf.random.uniform([], -1*self.saturation, self.saturation))
        return batch

    def __call__(self):
        return self.dataset().map(self.jitter)
    # todo:list nmax,arguments add



from .utils.models.segmentation_model import SegmentationModel
from .utils.data.augmentation import MultipleRandomRotation, MultipleColorJitter, MultipleToTensor
from .utils.data.dataloader import SegmentationDataLoader
import torch
import torch.nn as nn
from torch.utils.tensorboard import SummaryWriter
from torchvision import transforms
import os
from torch.utils.data import DataLoader
from os.path import join
from tqdm import tqdm
from time import localtime
from torch import optim

import matplotlib.pyplot as plt

from datetime import datetime

# TODO: Add types to args


default_transform = transforms.Compose([MultipleRandomRotation(10, fill=(255, 0)),
                                        MultipleColorJitter(
    brightness=0.3, contrast=0.3, n_max=1),
    MultipleToTensor(),
])


# General base class for Processors.
class BaseProcessor:
    def __init__(self, use_pretrained=True, lock_pretrained=True):
        self.model = SegmentationModel(use_pretrained, lock_pretrained)

    def load_data(self,
                  train_path,
                  val_path=None,
                  transform="default"):
        """
        Loads the data and creates a dataloader.

        Args:
            train_path (str): path to a directory containing two folders named
                `originals` and `labels` for training data.
            val_path (str, optional): path to a directory containing two
                named `originals` and `labels` for validation data.
                Default: None.
            transform (:mod:`~torchvision.transforms`, optional):
                Transforms to be applied on images and labels. It should be 
                able to take a list of images as inputs and return a list of
                transformed images.
        """

        if transform == "default":
            transform = default_transform

        self.train_dl = self._make_data(train_path, transform)
        self.val_dl = self._make_data(val_path, transform)

    def _make_data(self,
                   directory,
                   transform):
        if directory is not None:
            if "originals" in os.listdir(directory) and\
                    "labels" in os.listdir(directory):
                dataloader = SegmentationDataLoader(
                    join(directory, "originals"),
                    join(directory, "labels"),
                    transforms=transform
                )
            else:
                raise Exception(f"The path: {directory} does not contain\
                    'originals' or 'labels' directories.")

            return dataloader
        return None

    def train_step(self,
                   x,
                   y,
                   optimizer,
                   criterion,
                   ):

        optimizer.zero_grad()

        preds = self.model(x)
        loss = criterion(preds, y)
        loss.backward()
        optimizer.step()

        return loss.item()

    def val_step(self,
                 x,
                 y,
                 criterion):

        preds = self.model(x)

        loss = criterion(preds, y)

        return loss.item()

    def train_epoch(self,
                    epoch,
                    train_dl,
                    optimizer,
                    criterion,
                    writer,
                    train_running_loss,
                    log_freq,
                    device,):

        self.model.train()
        total_loss = 0

        # Progress bar
        pbar = tqdm(total=len(train_dl))
        pbar.set_description(f"Epoch: {epoch}. Traininig")

        for i, (x, y) in enumerate(train_dl, 1):

            x = x.to(device)
            y = y.to(device)

            loss = self.train_step(x, y, optimizer, criterion)
            total_loss += loss

            pbar.update()
            pbar.set_postfix(loss=total_loss/(i))

            train_running_loss += loss
            if ((epoch-1)*len(train_dl) + i) % log_freq == 0:

                writer.add_scalar('Loss/train',
                                  train_running_loss/log_freq,
                                  (epoch-1)*len(train_dl) + i)

                pred = self.predict(x[0].unsqueeze(0), device=device)
                _, index = pred.topk(1, dim=1)
                f, (ax1, ax2) = plt.subplots(1, 2)
                ax1.imshow(index[0, 0].cpu(), "gray")
                ax1.set_title("Prediction")
                ax2.imshow(y[0].cpu(), "gray")
                ax2.set_title("Ground Truth")

                writer.add_figure("Predictions/train",
                                  f,
                                  global_step=(epoch-1)*len(train_dl) + i)

                self.model.train()

                train_running_loss = 0.0

        pbar.close()

        return train_running_loss, total_loss

    def val_epoch(self,
                  epoch,
                  val_dl,
                  criterion,
                  writer,
                  val_running_loss,
                  log_freq,
                  device):
        self.model.eval()
        eval_loss = 0

        # validation progress bar
        pbar = tqdm(total=len(val_dl))
        pbar.set_description(f"Epoch: {epoch}. Validating")

        val_running_loss = 0
        for i, (x, y) in enumerate(val_dl, 1):

            x = x.to(device)
            y = y.to(device)

            loss = self.val_step(x, y, criterion)
            eval_loss += loss

            pbar.update(1)
            pbar.set_postfix(loss=eval_loss/(i))

            val_running_loss += loss
            if ((epoch-1)*len(val_dl) + i) % log_freq == 0:

                writer.add_scalar('Loss/val',
                                  val_running_loss/log_freq,
                                  (epoch-1)*len(val_dl) + i)

                pred = self.predict(x[0].unsqueeze(0), device=device)
                _, index = pred.topk(1, dim=1)
                f, (ax1, ax2) = plt.subplots(1, 2)
                ax1.imshow(index[0, 0].cpu(), "gray")
                ax1.set_title("Prediction")
                ax2.imshow(y[0].cpu(), "gray")
                ax2.set_title("Ground Truth")

                writer.add_figure("Predictions/val",
                                  f,
                                  global_step=(epoch-1)*len(val_dl) + i)

                val_running_loss = 0.0

        pbar.close()

        return val_running_loss, eval_loss

    def train(self,
              n_epochs,
              train_dl=None,
              val_dl=None,
              batch_size=1,
              shuffle=True,
              num_workers=4,
              pin_memory=True,
              criterion=None,
              optimizer=None,
              weight_decay=0,
              learning_rate=0.0001,
              learning_rate_decay=None,
              save_checkpoints=True,
              checkpoint_freq=None,
              checkpoint_path=None,
              max_to_keep=5,
              log_dir="./tensorboard_logs",
              log_freq=1000,
              device=None,
              ):

        # check if train loaders are ok
        if train_dl is None:

            # TEST IT

            if getattr(self, "train_dl", None) is None:
                raise Exception(
                    "No DataLoader found. Either pass one in train or use load_data method.")

            train_dl = self.train_dl(
                batch_size, shuffle, num_workers, pin_memory)
        if val_dl is None:
            if getattr(self, "val_dl", None) is None:
                val_dl = None
            else:
                val_dl = self.val_dl(
                    batch_size, shuffle, num_workers, pin_memory)
            ######

        if learning_rate_decay is None:
            learning_rate_decay = 1.0
        self.model.to(device)

        # checkpoint stuff
        if checkpoint_path is None:
            checkpoint_path = join(
                os.getcwd(), "checkpoints", self.__class__.__name__)
        os.makedirs(checkpoint_path, exist_ok=True)
        if checkpoint_freq is None:
            checkpoint_freq = n_epochs//10+1

        current_epoch = 0
        if len(os.listdir(checkpoint_path)) > 0:
            recent_file = self.get_latest_ckpt_path(
                checkpoint_path)

            checkpoint_file_path = join(checkpoint_path, recent_file)
            current_epoch, optimizer, loss = self.load_checkpoint(
                join(checkpoint_file_path), optimizer, device)

            print("Latest checkpoint found.")
            print(
                f"Epoch: {current_epoch}    loss: {loss}\nResuming training...")

        optimizer.param_groups[0]["lr"] = learning_rate
        optimizer.param_groups[0]["weight_decay"] = weight_decay
        scheduler = optim.lr_scheduler.ExponentialLR(
            optimizer, learning_rate_decay)

        current_epoch += 1

        now = datetime.now()
        train_running_loss = 0
        if val_dl is not None:
            val_running_loss = 0
        writer = SummaryWriter(join(log_dir, now.strftime("%Y%m%d-%H%M%S")))

        for epoch in range(current_epoch, current_epoch+n_epochs):

            train_running_loss, total_loss = self.train_epoch(
                epoch, train_dl, optimizer, criterion, writer, train_running_loss, log_freq, device)

            if val_dl is not None:
                val_running_loss, eval_loss = self.val_epoch(
                    epoch, val_dl, criterion, writer, val_running_loss, log_freq, device)

            scheduler.step()

            if epoch % checkpoint_freq == 0:
                name = "{}-{}-{}-{}-{}-{}-{}.pt".format(
                    epoch, *localtime()[:6])
                self.save_checkpoint(join(checkpoint_path, name), epoch,
                                     optimizer, total_loss/len(train_dl))

        writer.flush()
        writer.close()

    def save_checkpoint(self, path, epoch, optimizer, loss):
        r"""
        save the checkpoint (epoch, model parameters, optimizer parameters, loss).

        Args:
            path (str): path of the checkpoint file(.pt or .pth).
            epoch (int): epoch number to be saved in the checkpoint.
            optimizer (:class:`~torch.optim.Optimizer`): optimizer whose
                parameters need to be saved.
            loss (float): loss value to be stored.
        """
        checkpoint = {
            "epoch": epoch,
            "model": self.model.state_dict(),
            "optimizer": optimizer.state_dict(),
            "loss": loss,
        }
        torch.save(checkpoint, path)
        print("checkpoint saved")

    def load_checkpoint(self, path, optimizer, device):
        r"""
        loads a saved checkpoint.

        Args:
            path (str): path to a .pt or .pth file.
            optimizer (:class:`~torch.optim.Optimizer`): optimizer
                (preferably the same one which was saved) to load it's parameters.
            device (str): device to load on the checkpoint. "cuda" or "cpu".

        Returns:
            tuple(int, :class:`~torch.optim.Optimizer`, float): tuple of epoch,
            optimizer and loss.
        """
        checkpoint = torch.load(path)
        self.model.load_state_dict(checkpoint["model"])
        self.model.to(device)

        optimizer.load_state_dict(checkpoint["optimizer"])

        return checkpoint["epoch"], optimizer, checkpoint["loss"]

    def get_latest_ckpt_path(self, path):
        checkpoints = os.listdir(path)
        recent_time = datetime(
            *list(map(int, checkpoints[0].split(".")[0].split("-")))[1:])
        recent_file = checkpoints[0]
        for filename in checkpoints[1:]:
            date = list(map(int, filename.split(".")[0].split("-")))
            c_time = datetime(*date[1:])
            if c_time > recent_time:
                recent_time = c_time
                recent_file = filename
            elif c_time == recent_time:
                recent_file = filename if date[0] > int(
                    recent_file.split("-")[0]) else recent_file

        return recent_file

    def load_model(self, path):
        """
        Loads the model parameters for :attr:`self.model`.

        Args:
            path (str): path to a .pth(or .pt) file.
        """
        print(self.model.load_state_dict(torch.load(path)))
        print("Model Loaded!")

    def save_model(self, path):
        r"""saves the current model.

        Args:
            path(str): path to the file. (.pt or .pth)
        """
        if not (path.endswith(".pt") or path.endswith(".pth")):
            path = path+".pt"

        torch.save(self.model.state_dict(), path)
        print("Model Saved!")

    def predict(self, x, device=None):
        r"""predicts the output for a given input.

        Args:
            x (:class:`~torch.Tensor`): input tensor of shape 
                ``[batch_size, height, width, 3]``.
            device (str, optional): device on which to perform the computation.
                Default: GPU if it is available, else CPU.
        """
        # TODO: add dataloader option
        if device is None:
            device = torch.device(
                "cuda" if torch.cuda.is_available() else "cpu")

        self.model.to(device)
        x = x.to(device)
        self.model.eval()
        return self.model(x)


class PageExtractor(BaseProcessor):
    """
    This class is used to detect boundary of text in an image of a document.

    Args:
        use_pretrained (bool, optional): whether to use the parameters of pre-trained
            resnet50 model. 
            Default: ``True``.
        lock_pretrained (bool, optional): whether to lock the layers of pretrained
            model. Only needed if ``use_pretrained`` is ``True``.
            Default: ``True``.
    """

    def load_data(self,
                  train_path,
                  val_path=None,
                  transform="default"):
        """
        Loads the data and creates a dataloader.

        The label images should be:

        - the same name as the their original images (including the extension).
        - the same size as their original images.
        - black background with the page area filled with red color.

        Args:
            train_path (str): path to a directory containing two folders named
                `originals` and `labels` for training data.
            val_path (str, optional): path to a directory containing two
                named `originals` and `labels` for validation data.
                Default: None.
            transform (:mod:`~torchvision.transforms`, optional):
                Transforms to be applied on images and labels. It should be 
                able to take a list of images as inputs and return a list of
                transformed images.
        """
        super().load_data(train_path, val_path, transform)

    def train(self,
              n_epochs,
              train_dl=None,
              val_dl=None,
              batch_size=1,
              shuffle=True,
              num_workers=4,
              pin_memory=True,
              criterion=None,
              optimizer=None,
              weight_decay=0,
              learning_rate=0.0001,
              learning_rate_decay=None,
              save_checkpoints=True,
              checkpoint_freq=None,
              checkpoint_path=None,
              max_to_keep=5,
              log_dir="./tensorboard_logs",
              log_freq=1000,
              device=None,
              ):
        r"""
        Training method to detect boundary. 

        :attr:`train_dl` and :attr:`val_dl` should be provided if custom dataloaders
        are required. Although, most of the times, :meth:`~PageExtractor.load_data`
        will do the job. 

        Note:

            If :attr:`train_dl` and/or :attr:`val_dl` are provided, then any already made
            dataloader using  :meth:`~BaselineDetector.load_data` will not be used.

        Args:
            n_epochs (int): number of epochs.
            train_dl (:class:`~torch.utils.data.DataLoader`, optional): data loader to train on.
                Default: None.
            val_dl (:class:`~torch.utils.data.DataLoader`, optional): data loader to validate on.
                Default: None.
            batch_size (int, optional): batch size of the data loader created by :meth:`~BaselineDetector.load_data`.
                Default: 1.
            shuffle (bool, optional): whether to shuffle the data loader 
                created by :meth:`~BaselineDetector.load_data`.
            num_workers (int, optional): number of processes to load data.
                Default: 4
            pin_memory (bool, optional): whether to pin memory while loading data.
                Default: ``True``.
            criterion (:class:`~torch.nn.module.loss._Loss` or :class:`~torch.nn.module.loss._WeightedLoss`, optional): 
                The loss function to use.
                Default: ``CrossEntropyLoss(reduction="mean")``
            optimizer (:class:`~torch.optim.Optimizer`, optional): The optimizer
                to be used to update the parameters.
                Default: ``Adam(model.parameters(), lr=learning_rate)``
            weight_decay (float, optional): L2 regularization parameter for the optimizer(if it applies).
                Default: 0.
            learning_rate (float, optional): Learning rate to be used for the 
                default optimizer.
                Default: 0.0001.
            learning_rate_decay (float, optional): exponential decay to be used on learning rate.
                Default: None.
            save_checkpoints (bool, optional): Whether to save training checkpoints.
                Default: ``True``.
            checkpoint_freq (int, optional): The saving frequency. After each 
                these many epochs, a checkpoint will be saved.
                Default: :math:`\left \lfloor\frac{\text{n_epochs}}{10}\right \rfloor + 1`.
            checkpoint_path (str, optional): Path of the checkpoint folder.
                Default: "./"
            max_to_keep (int or None, optional): The maximum number of latest 
                checkpoints to be saved. ``None`` to save all. 
                Default: 5.
            device (str, optional): The device to do computations on.
                Default: GPU, if GPU is available, else CPU. 
        """

        if device is None:
            device = torch.device(
                "cuda" if torch.cuda.is_available() else "cpu")
        if optimizer is None:
            optimizer = optim.Adam(self.model.parameters(), lr=learning_rate)
        if criterion is None:
            criterion = nn.CrossEntropyLoss(reduction="mean")

        super().train(n_epochs, train_dl, val_dl, batch_size, shuffle,
                      num_workers, pin_memory, criterion, optimizer, weight_decay,
                      learning_rate, learning_rate_decay, save_checkpoints,
                      checkpoint_freq, checkpoint_path, max_to_keep, log_dir,
                      log_freq, device)


class ImageExtractor(BaseProcessor):
    """
    This class is used to detect images in a document.

    Args:
        use_pretrained (bool, optional): whether to use the parameters of pre-trained
            resnet50 model. 
            Default: ``True``.
        lock_pretrained (bool, optional): whether to lock the layers of pretrained
            model. Only needed if ``use_pretrained`` is ``True``.
            Default: ``True``.
    """

    def load_data(self,
                  train_path,
                  val_path=None,
                  transform="default"):
        """
        Loads the data and creates a dataloader.

        The label images should be:

        - the same name as the their original images (including the extension).
        - the same size as their original images.
        - black background with corresponding image area red. 

        Args:
            train_path (str): path to a directory containing two folders named
                `originals` and `labels` for training data.
            val_path (str, optional): path to a directory containing two
                named `originals` and `labels` for validation data.
                Default: None.
            transform (:mod:`~torchvision.transforms`, optional):
                Transforms to be applied on images and labels. It should be 
                able to take a list of images as inputs and return a list of
                transformed images.
        """
        super().load_data(train_path, val_path, transform)

    def train(self,
              n_epochs,
              train_dl=None,
              val_dl=None,
              batch_size=1,
              shuffle=True,
              num_workers=4,
              pin_memory=True,
              criterion=None,
              optimizer=None,
              weight_decay=0,
              learning_rate=0.0001,
              learning_rate_decay=None,
              save_checkpoints=True,
              checkpoint_freq=None,
              checkpoint_path=None,
              max_to_keep=5,
              log_dir="./tensorboard_logs",
              log_freq=1000,
              device=None,
              ):
        r"""
        Training method to detect images. 


        :attr:`train_dl` and :attr:`val_dl` should be provided if custom dataloaders
        are required. Although, most of the times, :meth:`~ImageExtractor.load_data`
        will do the job. 

        Note:

            If :attr:`train_dl` and/or :attr:`val_dl` are provided, then any already made
            dataloader using  :meth:`~BaselineDetector.load_data` will not be used.

        Args:
            n_epochs (int): number of epochs.
            train_dl (:class:`~torch.utils.data.DataLoader`, optional): data loader to train on.
                Default: None.
            val_dl (:class:`~torch.utils.data.DataLoader`, optional): data loader to validate on.
                Default: None.
            batch_size (int, optional): batch size of the data loader created by :meth:`~BaselineDetector.load_data`.
                Default: 1.
            shuffle (bool, optional): whether to shuffle the data loader 
                created by :meth:`~BaselineDetector.load_data`.
            num_workers (int, optional): number of processes to load data.
                Default: 4
            pin_memory (bool, optional): whether to pin memory while loading data.
                Default: ``True``.
            criterion (:class:`~torch.nn.module.loss._Loss` or :class:`~torch.nn.module.loss._WeightedLoss`, optional): 
                The loss function to use.
                Default: ``CrossEntropyLoss(reduction="mean")``
            optimizer (:class:`~torch.optim.Optimizer`, optional): The optimizer
                to be used to update the parameters.
                Default: ``Adam(model.parameters(), lr=learning_rate)``
            weight_decay (float, optional): L2 regularization parameter for the optimizer(if it applies).
                Default: 0.
            learning_rate (float, optional): Learning rate to be used for the 
                default optimizer.
                Default: 0.0001.
            learning_rate_decay (float, optional): exponential decay to be used on learning rate.
                Default: None.            
            save_checkpoints (bool, optional): Whether to save training checkpoints.
                Default: ``True``.
            checkpoint_freq (int, optional): The saving frequency. After each 
                these many epochs, a checkpoint will be saved.
                Default: :math:`\left \lfloor\frac{\text{n_epochs}}{10}\right \rfloor + 1`.
            checkpoint_path (str, optional): Path of the checkpoint folder.
                Default: "./"
            max_to_keep (int or None, optional): The maximum number of latest 
                checkpoints to be saved. ``None`` to save all. 
                Default: 5.
            device (str, optional): The device to do computations on.
                Default: GPU, if GPU is available, else CPU. 
        """

        if device is None:
            device = torch.device(
                "cuda" if torch.cuda.is_available() else "cpu")
        if optimizer is None:
            optimizer = optim.Adam(self.model.parameters(), lr=learning_rate)
        if criterion is None:
            criterion = nn.CrossEntropyLoss(reduction="mean")

        super().train(n_epochs, train_dl, val_dl, batch_size, shuffle,
                      num_workers, pin_memory, criterion, optimizer, weight_decay,
                      learning_rate, learning_rate_decay, save_checkpoints,
                      checkpoint_freq, checkpoint_path, max_to_keep, log_dir,
                      log_freq, device)


class BaselineDetector(BaseProcessor):
    """
    This class is used to detect baselines in an image of a document.

    Args:
        use_pretrained (bool, optional): whether to use the parameters of pre-trained
            resnet50 model. 
            Default: ``True``.
        lock_pretrained (bool, optional): whether to lock the layers of pretrained
            model. Only needed if ``use_pretrained`` is ``True``.
            Default: ``True``.
    """

    def load_data(self,
                  train_path,
                  val_path=None,
                  transform="default"):
        """
        Loads the data and creates a dataloader.

        The label images should be:

        - the same name as the their original images.
        - the same size as their original images.
        - black background with red lines of about 5px width representing baselines.


        Args:
            train_path (str): path to a directory containing two folders named
                `originals` and `labels` for training data.
            val_path (str, optional): path to a directory containing two
                named `originals` and `labels` for validation data.
                Default: None.
            transform (:mod:`~torchvision.transforms`, optional):
                Transforms to be applied on images and labels. It should be 
                able to take a list of images as inputs and return a list of
                transformed images.
        """
        super().load_data(train_path, val_path, transform)

    def train(self,
              n_epochs,
              train_dl=None,
              val_dl=None,
              batch_size=1,
              shuffle=True,
              num_workers=4,
              pin_memory=True,
              criterion=None,
              optimizer=None,
              weight_decay=0,
              learning_rate=0.0001,
              learning_rate_decay=None,
              save_checkpoints=True,
              checkpoint_freq=None,
              checkpoint_path=None,
              max_to_keep=5,
              log_dir="./tensorboard_logs",
              log_freq=1000,
              device=None,
              ):
        r"""
        Training method to detect baselines. 

        :attr:`train_dl` and :attr:`val_dl` should be provided if custom dataloaders
        are required. Although, most of the times, :meth:`~BaselineDetector.load_data`
        will do the job. 

        Note:

            If :attr:`train_dl` and/or :attr:`val_dl` are provided, then any already made
            dataloader using  :meth:`~BaselineDetector.load_data` will not be used.

        Args:
            n_epochs (int): number of epochs.
            train_dl (:class:`~torch.utils.data.DataLoader`, optional): data loader to train on.
                Default: None.
            val_dl (:class:`~torch.utils.data.DataLoader`, optional): data loader to validate on.
                Default: None.
            batch_size (int, optional): batch size of the data loader created by :meth:`~BaselineDetector.load_data`.
                Default: 1.
            shuffle (bool, optional): whether to shuffle the data loader 
                created by :meth:`~BaselineDetector.load_data`.
            num_workers (int, optional): number of processes to load data.
                Default: 4
            pin_memory (bool, optional): whether to pin memory while loading data.
                Default: ``True``.
            criterion (:class:`~torch.nn.module.loss._Loss` or :class:`~torch.nn.module.loss._WeightedLoss`, optional): 
                The loss function to use.
                Default: ``CrossEntropyLoss(weight=Tensor[1, 700]), reduction="mean")``
            optimizer (:class:`~torch.optim.Optimizer`, optional): The optimizer
                to be used to update the parameters.
                Default: ``Adam(model.parameters(), lr=learning_rate)``
            weight_decay (float, optional): L2 regularization parameter for the optimizer(if it applies).
                Default: 0.
            learning_rate_decay (float, optional): exponential decay to be used on learning rate.
                Default: None.
            learning_rate (float, optional): Learning rate to be used for the 
                default optimizer.
                Default: 0.0001.
            save_checkpoints (bool, optional): Whether to save training checkpoints.
                Default: ``True``.
            checkpoint_freq (int, optional): The saving frequency. After each 
                these many epochs, a checkpoint will be saved.
                Default: :math:`\left \lfloor\frac{\text{n_epochs}}{10}\right \rfloor + 1`.
            checkpoint_path (str, optional): Path of the checkpoint folder.
                Default: "./"
            max_to_keep (int or None, optional): The maximum number of latest 
                checkpoints to be saved. ``None`` to save all. 
                Default: 5.
            device (str, optional): The device to do computations on.
                Default: GPU, if GPU is available, else CPU. 
        """

        if device is None:
            device = torch.device(
                "cuda" if torch.cuda.is_available() else "cpu")
        if optimizer is None:
            optimizer = optim.Adam(self.model.parameters(), lr=learning_rate)
        if criterion is None:
            criterion = nn.CrossEntropyLoss(
                weight=torch.Tensor([1, 700]).to(device), reduction="mean")

        super().train(n_epochs, train_dl, val_dl, batch_size, shuffle,
                      num_workers, pin_memory, criterion, optimizer, weight_decay,
                      learning_rate, learning_rate_decay, save_checkpoints,
                      checkpoint_freq, checkpoint_path, max_to_keep, log_dir,
                      log_freq, device)



import tensorflow as tf
from tensorflow.keras import layers, Model


class ConvBlock(Model):
    r""" This block consists of five :class:`tensorflow.keras.layers.Conv2D` convolution layers and a :class:`tensorflow.keras.layers.MaxPool2D` maxpool layer.
        Convolution layers is of  kernel size  :math:`3` ,same padding and relu activation, and filters :math:`8 , 16 , 32 , 64` and :math:`128` respectively.
        Maxpool layer is of pool size :math:`(2 , 2)`, strides :math:`( 2 , 2)` ans same padding.


    """

    def __init__(self):
        super(ConvBlock, self).__init__()
        self.conv1 = layers.Conv2D(
            filters=8, kernel_size=3, padding='same', activation='relu')
        self.conv2 = layers.Conv2D(
            filters=16, kernel_size=3, padding='same', activation='relu')
        self.conv3 = layers.Conv2D(
            filters=32, kernel_size=3, padding='same', activation='relu')
        self.conv4 = layers.Conv2D(
            filters=64, kernel_size=3, padding='same', activation='relu')
        self.conv5 = layers.Conv2D(
            filters=128, kernel_size=3, padding='same', activation='relu')
        self.pool1 = layers.MaxPool2D(pool_size=(
            2, 2), strides=(2, 2), padding='same')

    def call(self, input_tensor):
        r"""

        Args:
            input_tensor ( :class:`tensorflow.Tensor`) : Input image tensor.

        Returns:
            :class:`tensorflow.Tensor` : Output image tensor.
        """
        x = self.conv1(input_tensor)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        x = self.conv5(x)
        x = self.pool1(x)
        return x


class BLSTM(Model):
    r"""Consists of a two :class:`tensorflow.keras.layers.Bidirectional` layer with :math:`256` units.

    """

    def __init__(self):
        super(BLSTM, self).__init__()

        self.layer1 = layers.Bidirectional(
            layers.LSTM(units=256, return_sequences=True))
        self.layer2 = layers.Bidirectional(
            layers.LSTM(units=256, return_sequences=True))

    def call(self, input_tensor):
        r"""
        Args:
            input_tensor( :class:`tensorflow.Tensor`): Input tensor.

        Returns:
            :class:`tensorflow.Tensor` : Output tensor.
        """
        x = self.layer1(input_tensor)
        x = self.layer2(x)
        return x


class OCRModel(Model):
    r"""Consistis of :class:`~batukh.tensorflow.utils.models.ocr_model.ConvBlock` block, :class:`~batukh.tensorflow.utils.models.ocr_model.BLSTM` block and :class:`tensorflow.keras.layers.dense` layer.

    Args:
        n_classes (int)  : Specifies number of output classes in :class:`tensorflow.keras.layers.dense` layer.
    """

    def __init__(self, n_classes):
        super(OCRModel, self).__init__()
        self.n_classes = n_classes
        self.conv = ConvBlock()
        self.blstm = BLSTM()
        self.dense = layers.Dense(units=self.n_classes)

    def call(self, input_tensor):
        r"""
        Args:
            input_tensor ( :class:`tensorflow.Tensor`) : Input image tensor.

        Returns:
            :class:`tensorflow.Tensor` : Output Tensor."""
        x = self.conv.call(input_tensor)
        x = tf.transpose(x, (0, 2, 1, 3))
        x = layers.Reshape((-1, x.shape[-2]*128))(x)
        x = self.blstm.call(x)
        x = self.dense(x)
        return x



import tensorflow as tf
import time
from tqdm import tqdm
import numpy as np
import tensorflow_addons as tfa
import matplotlib.pyplot as plt
import io


class Train():
    r"""Contains methods to train a model and get predictions.

    Args:
        model           : Model to be trained.
        is_ocr (bolean) : Specifies if model is :class:`~batukh.tensorflow.utils.models.ocr_model.OCR`.
            Default : ``False``
    """

    def __init__(self, model, is_ocr=False):
        self.is_ocr = is_ocr
        self.model = model

        localtime = time.asctime()

        self.train_summary_writer = tf.summary.create_file_writer(
            f"logs/{localtime}/train")
        self.val_summary_writer = tf.summary.create_file_writer(
            f"logs/{localtime}/val")
        self.train_loss = tf.keras.metrics.Mean(name="loss", dtype=tf.float32)
        self.val_loss = tf.keras.metrics.Mean(name="loss", dtype=tf.float32)

    def _train_one_step(self, x, y):
        r"""Trains one image or batch of image.

        Args:
            x ( :class:`tensorflow.Tensor`) : Input image tensor or batch of image tensor.
            y ( :class:`tensorflow.Tensor`) : Target image tensor or batch of image tensor.

        Returns:
            :class:`tensorflow.float32` : loss.

        """
        with tf.GradientTape() as tape:
            logits = self.model(x, training=True)
            if self.is_ocr:
                logit_length = tf.fill(
                    [tf.shape(logits)[0]], tf.shape(logits)[1])
                loss = tf.nn.ctc_loss(labels=y, logits=logits, label_length=None,
                                      logit_length=logit_length, logits_time_major=False, blank_index=0)
                loss = tf.reduce_mean(loss)
            else:
                loss = self.criterion(y, logits)
                loss, _ = tf.nn.weighted_moments(
                    loss, (1, 2), np.sum(y*self.class_weights, axis=3))
                loss = loss[0]

        grads = tape.gradient(loss, self.model.trainable_variables)
        self.optimizer.apply_gradients(
            zip(grads, self.model.trainable_variables))
        return loss

    def _train(self, ds, epoch, batch_size, repeat, log_freq):
        r"""
        Args:
            ds ( :class:`tensorflow.data.Dataset`)  : dataset.
            epoch (int)      : epoch.
            batch_size (int) : Batch size of dataloader.
            repeat (int)     : Number of times dataloader will be itterated.

        """
        pbar = tqdm(total=len(ds)*repeat)
        pbar.set_description(f"Epoch: {epoch}. Traininig")

        for x, y in ds(batch_size, repeat):
            loss = self._train_one_step(x, y)
            self.train_loss.update_state(loss)
            if tf.equal(self.optimizer.iterations % log_freq, 0):
                img = tf.math.argmax(self.predict(x)[0], -1)
                with self.train_summary_writer.as_default():
                    tf.summary.scalar('loss', self.train_loss.result(),
                                      step=self.optimizer.iterations)

                    tf.summary.image("Predictions/train",
                                     self._plot_to_image(
                                         self._plot(img, y[0][:, :, 1])),
                                     step=self.optimizer.iterations)

            pbar.update(1)
            pbar.set_postfix(loss=float(self.train_loss.result()))
        pbar.close()

    def train(self, n_epochs, train_dl=None, val_dl=None,  batch_size=1, repeat=1, criterion=None, class_weights=None, optimizer=None, weight_decay=None, learning_rate=0.0001, lr_decay=None, save_checkpoints=True, checkpoint_freq=None, checkpoint_path=None, max_to_keep=5, log_freq=100):
        r"""
        Args:
            n_epochs (int) : Number of epochs.
            train_dl ( :class:`tensorflow.data.Dataset`,optional) : Dataset for training.
                Default: ``None``
            val_dl ( :class:`tensorflow.data.Dataset`,optional): Dataset for validation.
                Default: ``None``
            batch_size (int,optional) : batch size.
                Default: :math:`1`
            repeat (int,optional) : specifies the number of times dataset will be itterated in one epoch.
                Default : :math:`1`
            criterion (optional): Specifies loss function to be used in training.
                Default : if is_ocr is ``True`` :class:`tensorflow.nn.ctc_loss` used
                else :class:`tensorflow.nn.softmax_cross_entropy_with_logits` used.
            class_weights (list,,optional) : weights given to class while calculating loss.
                Default : Same for each class.
            optimizer (optional) : optimizer used in training.
                Default : :class:`tensorflow.keras.optimizers.Adam`
            learning_rate (float,optional):learning rate of optimizer.
                Default: :math:`0.0001`
            save_checkpoints (bol,optional) : Specifes whether to save checkpoints.
                Default: ``True``
            checkpoint_freq (int,optional) : Specifies the number of epochs after checkpoints will be saved.
                Default:  :math:`=\lfloor \frac{ \text{n_epochs}}{10}+1 \rfloor`
            checkpoint_path (str,optional): Path where  checkpoints will be saved or loaded from.
                Default :  `"tf_ckpts/{}".format(` :class:`time.asctime` `)`
            max_to_keep (int,optional) : Maximun number of latest checkpoints to keep.
                Default : :math:`5`
                """
        if lr_decay is not None:
            learning_rate = tf.keras.optimizers.schedules.ExponentialDecay(
                learning_rate, decay_steps=100000, decay_rate=lr_decay, staircase=True)

        if optimizer is None:
            optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate)
        if weight_decay is not None:
            optimizer = tfa.optimizers.AdamW(
                learning_rate=learning_rate, weight_decay=weight_decay)
        self.optimizer = optimizer
        if criterion is None:
            if self.is_ocr:
                criterion = tf.nn.ctc_loss
            else:
                criterion = tf.nn.softmax_cross_entropy_with_logits
        self.criterion = criterion
        if class_weights is None:
            class_weights = [1]*self.model.n_classes
        self.class_weights = class_weights
        localtime = time.asctime()

        if checkpoint_path is None:
            checkpoint_path = "tf_ckpts/{}".format(localtime)
        self.checkpoint = tf.train.Checkpoint(
            model=self.model, optimizer=self.optimizer)

        if checkpoint_freq is None:
            checkpoint_freq = n_epochs//10 + 1

        self.manager = tf.train.CheckpointManager(
            self.checkpoint,
            directory=checkpoint_path,
            max_to_keep=max_to_keep)
        self.checkpoint.restore(self.manager.latest_checkpoint)

        if self.manager.latest_checkpoint:
            print("Restored from {}".format(self.manager.latest_checkpoint))
        else:
            print("Initializing from scratch")

        for epoch in range(1, n_epochs + 1):
            self._train(train_dl, epoch, batch_size, repeat, log_freq)
            if epoch % checkpoint_freq == 0:
                checkpoint_path = self.manager.save(self.optimizer.iterations)
                print("Model saved to {}".format(checkpoint_path))
            if val_dl is not None:
                with self.val_summary_writer.as_default():
                    self._val(val_dl,  batch_size, repeat, epoch, log_freq)
            self.val_loss.reset_states()
            self.train_loss.reset_states()

    def _val_one_step(self, x, y):
        r"""

        Args:
            x ( :class:`tensorflow.Tensor`) : input image tensor.
            y ( :class:`tensorflow.Tensor`) : input label tensor.

        Returns:
            :class:`tensorflow.float` :loss
            :class:`tensorflow.Tensor` : logits
            """
        logits = self.model(x, training=False)
        if self.is_ocr:
            logit_length = tf.fill(
                [tf.shape(logits)[0]], tf.shape(logits)[1])
            loss = tf.nn.ctc_loss(labels=y, logits=logits, label_length=None,
                                  logit_length=logit_length, logits_time_major=False, blank_index=0)
            loss = tf.reduce_mean(loss)
        else:
            loss = self.criterion(y, logits)
            loss, _ = tf.nn.weighted_moments(
                loss, (1, 2), np.sum(y*self.class_weights, axis=3))
            loss = loss[0]

        return loss, logits

    def _val(self, ds, epoch,  batch_size, repeat, log_freq):
        r"""
        Args:
            ds ( :class:`tensorflow.data.Dataset`) : datsset.
            epoch (int) : epoch.
            batch_size (int) : batch size.
            repeat (int): number of times dataset will be itterated in one epoch.
            """
        pbar = tqdm(total=len(ds)*repeat)
        pbar.set_description(f"Epoch: {epoch}. validation")
        for i, (x, y) in enumerate(ds(batch_size, repeat)):
            loss, _ = self._val_one_step(x, y)
            self.val_loss.update_state(loss)
            if tf.equal(((epoch-1)*len(ds)+i) % log_freq, 0):
                img = tf.math.argmax(self.predict(x)[0], -1)
                with self.train_summary_writer.as_default():
                    tf.summary.scalar('loss', self.val_loss.result(),
                                      step=self.optimizer.iterations)
                    tf.summary.image("Predictions/val",
                                     self._plot_to_image(
                                         self._plot(img, y[0][:, :, 1])),
                                     step=self.optimizer.iterations)
            pbar.update(1)
            pbar.set_postfix(loss=float(self.val_loss.result()))
        pbar.close()

    def predict(self, x):
        r"""
        Args:
            x ( :class:`tensorflow.Tensor`) : Input image tensor.

        Returns:
            :class:`tensorflow.Tensor` : predictions.
            """

        logits = self.model(x, training=False)
        return logits

    def save_model(self, path, name):
        r"""Saves model

        Args:
            path (str)          : path of the folder where model is to be saved.
            """

        self.model.save_weights(path)
        print("Model saved at "+path)

    def load_model(self, path):
        r"""Loads presaved weights.

        Args:
            path (str) : path of the model to be loaded.

        """
        self.model.load_weights(path)
        print("Model Loaded")

    def _plot_to_image(self, figure):
        """
        Converts the matplotlib plot specified by 'figure' to a PNG image and
        returns it. The supplied figure is closed and inaccessible after this call.
        """
        buf = io.BytesIO()
        plt.savefig(buf, format='png')
        plt.close(figure)
        buf.seek(0)
        image = tf.image.decode_png(buf.getvalue(), channels=4)
        image = tf.expand_dims(image, 0)
        return image

    def _plot(self, x, y):
        fig, (ax1, ax2) = plt.subplots(1, 2)
        ax1.imshow(x, "gray")
        ax1.set_title("Prediction")
        ax2.imshow(y, "gray")
        ax2.set_title("Ground Truth")
        return fig



from torchvision import transforms
import torchvision.transforms.functional as TF
from PIL import Image


class MultipleRandomRotation(transforms.RandomRotation):
    """
    Randomly rotates a list of images by the same degrees.

    Args:
        degrees (tuple(int, int) or int) : The range from whuch a random
            orientation is chosen. If ``int``, then range is ``(-degrees, degrees)``.
        resample (bool, optional): Whether or not to resample the images. Default: ``False``.
        expand (bool, optional): Whether or not to expand the images. Default: ``Flase``.
        center(): 
        fill (tuple(int, int, int) or int, optional): The color value to be filled at new places. 
    """

    def __init__(self,
                 degrees,
                 resample=False,
                 expand=False,
                 center=None,
                 fill=None):
        super(MultipleRandomRotation, self).__init__(
            degrees, resample, expand, center, fill)

    def __call__(self, images):

        if self.fill is None:
            self.fill = [None]*len(images)

        angle = self.get_params(self.degrees)
        return [TF.rotate(img, angle, self.resample, self.expand, self.center,
                          self.fill[i]) for i, img in enumerate(images)]


class MultipleColorJitter(transforms.ColorJitter):
    """
    Randomly jitter a list of images.

    Args:
        n_max (int, optional): If set then first ``n_max`` images in the list 
            will be applied the same jitter. Others will remain unchanged.
            Default: ``len(images)``
    """

    def __init__(self, brightness=0, contrast=0, saturation=0, hue=0, n_max=None):
        super(MultipleColorJitter, self).__init__(
            brightness, contrast, saturation, hue)

        self.n_max = n_max

    def __call__(self, images):
        transform = self.get_params(self.brightness, self.contrast,
                                    self.saturation, self.hue)

        if self.n_max is None:
            self.n_max = len(images)

        out = [transform(images[i]) for i in range(self.n_max)]
        out.extend(images[self.n_max:])
        return out


class MultipleToTensor(transforms.ToTensor):
    def __init__(self):
        super(MultipleToTensor, self).__init__()

    def __call__(self, images):
        return [TF.to_tensor(img) for img in images]



import tensorflow as tf
from tensorflow.keras import layers, Model


class BottleneckUnit(Model):
    r"""Consists of :math:`3` :class:`tensorflow.keras.layers.Conv2D` convolution layers with same padding,
    relu activation and kernel size :math:`1,3,1` respectively.

    Args:
        out_filters (int)         : Specifies filters of third convulution layer.
        in_filters (int)          : Specifies fliters of first and second convulution layer.
        strides (tuple,optional)  : Specifies strides of convulution layers.
            Default : :math:`(1,1)`
        connect (bol,optional) : Specifies if the layer will be used for horizantal connection in upscaler.
            Default :``True``.
        is_first (bol,optional): Specifies if the bottleneck unit is first of the bollteneck block.
            Default :``False``.


    """

    def __init__(self, out_filters, in_filters, stride=(1, 1), connect=True, is_first=False):
        super(BottleneckUnit, self).__init__()

        self.connect = connect
        self.is_first = is_first

        self.conv1 = layers.Conv2D(
            filters=in_filters, kernel_size=1, strides=(
                1, 1), activation='relu')
        self.conv2 = layers.Conv2D(
            filters=in_filters, kernel_size=3, strides=stride, padding="same", activation='relu')
        self.conv3 = layers.Conv2D(
            filters=out_filters, kernel_size=1, strides=(1, 1))
        self.conv4 = layers.Conv2D(
            filters=out_filters, kernel_size=1, strides=(1, 1))

    def call(self, input_tensor):
        r"""

        Args:
            input_tensor ( :class:`tensorflow.Tensor`) : Input image tensor.

        Returns:
            :class:`tensorflow.Tensor` : Output image tensor.
        """
        x = self.conv1(input_tensor)
        x = self.conv2(x)
        x = self.conv3(x)
        if self.connect:
            if self.is_first:
                input_tensor = self.conv4(input_tensor)
            return x + input_tensor
        return x


class ResnetLayer(Model):
    r"""Consists of three :class:`tensorflow.keras.layers.conv2D` convolution layer, 
    :class:`tensorflow.keras.layers.MaxPool2D` maxpoll layer and :math:`16`  
    :class:`~batukh.tensorflow.utils.models.segmentation_model.BottleneckUnit` units.

    """

    def __init__(self):
        super(ResnetLayer, self).__init__()

        self.conv1 = layers.Conv2D(
            filters=64, kernel_size=7, strides=(
                2, 2), padding="same")
        self.pool = layers.MaxPool2D(strides=(2, 2))

        self.bottleneck1 = BottleneckUnit(256, 64, is_first=True)
        self.bottleneck2 = BottleneckUnit(256, 64)
        self.bottleneck3 = BottleneckUnit(
            256, 64, stride=(2, 2), connect=False)

        self.bottleneck4 = BottleneckUnit(512, 128, is_first=True)
        self.bottleneck5 = BottleneckUnit(512, 128)
        self.bottleneck6 = BottleneckUnit(512, 128)
        self.bottleneck7 = BottleneckUnit(512, 128, (2, 2), connect=False)

        self.bottleneck8 = BottleneckUnit(1024, 256, is_first=True)
        self.bottleneck9 = BottleneckUnit(1024, 256)
        self.bottleneck10 = BottleneckUnit(1024, 256)
        self.bottleneck11 = BottleneckUnit(1024, 256)
        self.bottleneck12 = BottleneckUnit(1024, 256)
        self.bottleneck13 = BottleneckUnit(1024, 256, (2, 2), connect=False)

        self.bottleneck14 = BottleneckUnit(2048, 512, is_first=True)
        self.bottleneck15 = BottleneckUnit(2048, 512)
        self.bottleneck16 = BottleneckUnit(2048, 512)

        self.conv2 = layers.Conv2D(
            512, kernel_size=1, strides=(
                1, 1), padding="same")
        self.conv3 = layers.Conv2D(
            512, kernel_size=1, strides=(
                1, 1), padding="same")

    def call(self, input_tensor):
        r"""
        Args:
            input_tensor ( :class:`tensorflow.Tensor`) : Input image tensor.

        Returns:
            :class:`tensorflow.Tensor`        : Output image tensor.
            list : List of image tensor used in horizantal connections of upscaller.
        """
        concat_tensors = []
        concat_tensors.append(input_tensor)
        x = self.conv1(input_tensor)
        concat_tensors.append(x)

        x = self.pool(x)

        x = self.bottleneck1.call(x)
        x = self.bottleneck2.call(x)
        concat_tensors.append(x)
        x = self.bottleneck3.call(x)
        x = self.bottleneck4.call(x)
        x = self.bottleneck5.call(x)
        x = self.bottleneck6.call(x)
        concat_tensors.append(x)
        x = self.bottleneck7.call(x)
        x = self.bottleneck8.call(x)
        x = self.bottleneck9.call(x)
        x = self.bottleneck10.call(x)
        x = self.bottleneck11.call(x)
        x = self.bottleneck12.call(x)

        y = self.conv2(x)
        concat_tensors.append(y)

        x = self.bottleneck13.call(x)
        x = self.bottleneck14.call(x)
        x = self.bottleneck15.call(x)
        x = self.bottleneck16.call(x)

        x = self.conv3(x)
        return x, concat_tensors


class UpScalerUnit(Model):
    r"""Consists of :class:`layers.Conv2DT` convolution transpose layer and
    :class:`tensorflow.keras.layers.Conv2D` convolution layer.

    Args:
        in_filters (int)  : Specifies the filters of convolution transpose layer.
        out_filters (int) : Specifies the filters of convolution layer.

    """

    def __init__(self, in_filters, out_filters):
        super(UpScalerUnit, self).__init__()

        self.convT = layers.Conv2DTranspose(
            in_filters, kernel_size=1, strides=(2, 2))
        self.conv = layers.Conv2D(
            out_filters,
            kernel_size=3,
            padding='same',
            activation='relu')

    def call(self, input_tensor, concat_tensor):
        r"""
        Args:
            input_tensor ( :class:`tensorflow.Tensor`) : Input image tensor.
            concat_tensor ( :class:`tensorflow.Tensor`): Image tensor used to concat with the input tensor.

        Returns:
            :class:`tensorflow.Tensor` : Output image tensor.
        """
        x = self.convT(input_tensor)
        x = tf.concat([concat_tensor, x], 3)
        x = self.conv(x)
        return x


class UpScaler(Model):
    r"""Consists of :class:`tensorflow.keras.layers.Conv2D` convolution layer and 
    five :class:`~batukh.tensorflow.utils.models.segmentation_model.UpScalerUnit`.

    Args:
        n_classes (int) : Specifies the number of classes or channels in output image.
    """

    def __init__(self, n_classes):
        super(UpScaler, self).__init__()

        self.upScaler1 = UpScalerUnit(512, 512)
        self.upScaler2 = UpScalerUnit(512, 256)
        self.upScaler3 = UpScalerUnit(256, 128)
        self.upScaler4 = UpScalerUnit(128, 64)
        self.upScaler5 = UpScalerUnit(64, 32)
        self.conv = layers.Conv2D(
            filters=n_classes,
            kernel_size=1,
            padding='same')

    def call(self, input_tensor, concat_tensors):
        r"""
        Args:
            input_tensor ( :class:`tensorflow.Tensor`) : Input image tensor.
            concat_tensors (list)    : List of concat tensors used for horizantal connections.

        Returns:
            :class:`tensorflow.Tensor` : Output image tensor with channels equal to n_classes.
            """
        x = self.upScaler1.call(input_tensor, concat_tensors.pop(-1))
        x = self.upScaler2.call(x, concat_tensors.pop(-1))
        x = self.upScaler3.call(x, concat_tensors.pop(-1))
        x = self.upScaler4.call(x, concat_tensors.pop(-1))
        x = self.upScaler5.call(x, concat_tensors.pop(-1))
        x = self.conv(x)
        return x


class SegmentationModel(Model):
    r"""Consists of :class:`~batukh.tensorflow.utils.models.segmentation_model.ResnetLayer` and  
    :class:`~batukh.tensorflow.utils.models.segmentation_model.UpScaler`.

    Args:
        n_classes (int) : Specifies number of classes or channels in output image tensor.

    """

    def __init__(self, n_classes):
        super(SegmentationModel, self).__init__()
        self.n_classes = n_classes
        self.resnet = ResnetLayer()
        self.upScaler = UpScaler(self.n_classes)

    def call(self, input_tensor):
        r"""
        Args:
            input_tensor ( :class:`tensorflow.Tensor`) : Input image tensor.

        Returns:
            :class:`tensorflow.Tensor` : Output image tensor with channels equal to n_classes.
        """
        x, concat_tensors = self.resnet.call(input_tensor)
        x = self.upScaler.call(x, concat_tensors)
        return x



from torchvision.models.resnet import ResNet, Bottleneck
from torchvision import models
import torch.nn as nn
import torch
import torch.nn.functional as F
import os
from os.path import join

# Data Prep

# Remove this??
import numpy as np


# Transforms


# Models


class Identity(nn.Module):
    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, x):
        return x


class ModifiedBottleneck(Bottleneck):
    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None, is_last=False):

        super(ModifiedBottleneck, self).__init__(inplanes, planes, stride, downsample, groups,
                                                 base_width, dilation, norm_layer)

        self.is_last = is_last

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        if not self.is_last:
            out += identity
        out = self.relu(out)

        return out


class ResNetLayer(nn.Module):
    def __init__(self, block, inplanes, planes, blocks, sec_output_block=None, strides=None, connect_last=False):
        super(ResNetLayer, self).__init__()

        if strides is None:
            strides = [1]*blocks

        self.blocks = blocks
        self.sec_output_block = sec_output_block
        planes = int(planes/block.expansion)

        downsample = nn.Sequential(
            nn.Conv2d(inplanes, planes * block.expansion,
                      kernel_size=(1, 1), stride=(1, 1), bias=False),
            nn.BatchNorm2d(planes * block.expansion, eps=1e-05,
                           momentum=0.1, affine=True, track_running_stats=True)
        )

        self.add_module("0", block(inplanes, planes, strides[0], downsample))
        inplanes = int(planes * block.expansion)

        for i in range(1, blocks-1):
            self.add_module(str(i), block(inplanes, planes, strides[i]))
        i += 1
        if connect_last:
            is_last = False
        else:
            is_last = True
        self.add_module(str(i), block(
            inplanes, planes, strides[i], is_last=is_last))

    def forward(self, x):

        for i in range(self.blocks):
            block = getattr(self, str(i))
            x = block(x)
            if i == self.sec_output_block:
                sec_output = x.clone()
        if self.sec_output_block is not None:
            return x, sec_output
        return x


class MyResNet(nn.Module):
    def __init__(self, block, layers, strides=None, norm_layer=None, sec_output_blocks=None):
        super(MyResNet, self).__init__()

        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if strides is None:
            strides[None]*len(layers)
        if sec_output_blocks is None:
            sec_output_blocks = [None]*layers

        self.inplanes = 64
        self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=7, stride=2, padding=3,
                               bias=False)
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.layer1 = ResNetLayer(block, self.inplanes, self.inplanes*block.expansion, layers[0],
                                  strides=strides[0], sec_output_block=sec_output_blocks[0])

        self.inplanes = self.inplanes*block.expansion

        self.layer2 = ResNetLayer(block, self.inplanes, self.inplanes*block.expansion//2, layers[1],
                                  strides=strides[1], sec_output_block=sec_output_blocks[1])

        self.inplanes = self.inplanes*block.expansion//2

        self.layer3 = ResNetLayer(block, self.inplanes, self.inplanes*block.expansion//2, layers[2],
                                  strides=strides[2], sec_output_block=sec_output_blocks[2])

        self.inplanes = self.inplanes*block.expansion//2

        self.layer4 = ResNetLayer(block, self.inplanes, self.inplanes*block.expansion//2, layers[3],
                                  strides=strides[3], sec_output_block=sec_output_blocks[3])

    def forward(self, x):
        sec_outputs = []

        sec_outputs.append(x.clone())

        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        sec_outputs.append(x.clone())

        x = self.maxpool(x)

        x, sec = self.layer1(x)
        sec_outputs.append(sec)

        x, sec = self.layer2(x)
        sec_outputs.append(sec)

        x, sec = self.layer3(x)
        sec_outputs.append(sec)

        x = self.layer4(x)

        return x, sec_outputs


class UpScalerUnit(nn.Module):
    def __init__(self, n_current, n_output, n_copy):
        super(UpScalerUnit, self).__init__()
        # self.upsample = nn.UpsamplingNearest2d(scale_factor=2)
        self.conv = nn.Conv2d(n_current+n_copy, n_output,
                              kernel_size=(3, 3), padding=1, bias=False)

    def forward(self, x, copy):
        # print(x.shape, copy.shape)

        # Dynamically created a layer to match the size
        self.upsample = nn.UpsamplingNearest2d(
            size=(copy.shape[-2], copy.shape[-1]))
        x = self.upsample(x)
        x = torch.cat((copy, x), dim=1)
        x = self.conv(x)
        return x


class UpScaler(nn.Module):
    def __init__(self, n_classes):
        super(UpScaler, self).__init__()

        self.upscale1 = UpScalerUnit(512, 512, 512)

        self.upscale2 = UpScalerUnit(512, 256, 512)

        self.upscale3 = UpScalerUnit(256, 128, 256)

        self.upscale4 = UpScalerUnit(128, 64, 64)

        self.upscale5 = UpScalerUnit(64, 32, 3)

        self.conv2 = nn.Conv2d(32, n_classes, kernel_size=1, bias=False)

    def forward(self, x, copies):

        x = self.upscale1(x, copies.pop(-1))
        x = self.upscale2(x, copies.pop(-1))
        x = self.upscale3(x, copies.pop(-1))
        x = self.upscale4(x, copies.pop(-1))
        x = self.upscale5(x, copies.pop(-1))

        x = self.conv2(x)

        return x


class SegmentationModel(nn.Module):
    def __init__(self, use_pretrained=True, lock_pretrained=True):
        super(SegmentationModel, self).__init__()

        self.downsampler = MyResNet(ModifiedBottleneck, [3, 4, 6, 3],
                                    strides=[[1, 1, 2], [1, 1, 1, 2],
                                             [1, 1, 1, 1, 1, 2], None],
                                    sec_output_blocks=[1, 2, 4, None])

        if use_pretrained:
            resnet = models.resnet50(pretrained=True)
            resnet.avgpool = Identity()
            resnet.fc = Identity()
            print(self.downsampler.load_state_dict(resnet.state_dict()))

            if lock_pretrained:
                for param in self.downsampler.parameters():
                    param.requires_grad = False

        self.conv1 = nn.Conv2d(1024, 512, kernel_size=(1, 1), bias=False)
        self.conv2 = nn.Conv2d(2048, 512, kernel_size=(1, 1), bias=False)
        self.upscaler = UpScaler(2)

    def forward(self, x):
        x, sec_outputs = self.downsampler(x)
        sec_outputs[-1] = self.conv1(sec_outputs[-1])
        x = self.conv2(x)
        x = self.upscaler(x, sec_outputs)

        return x



import os
import tensorflow as tf
import numpy as np

# todo: test ocrdataloader


class SegmentationDataLoader():
    r""" Loads the :class:`tensorflow.data.Dataset` for :class:`~batukh.tensorflow.segmenter.PageExtractor`, 
    :class:`~batukh.tensorflow.segmenter.ImageExtractor`, :class:`~batukh.tensorflow.segmenter.LayoutExtractor` and 
    :class:`~batukh.tensorflow.segmenter.BaselineDetector` classes.

    Example

    .. code:: python

        >>> from batukh.tensorflow.utils.data.dataloader import SegmentationDataLoader
        >>> dl=SegmentationDataLoader("/data/",2)
        >>> for i,j in dl(batch_size=1):
        ...     print(i.shape,j.shape)
        ...     break

    Args:
        path (str)       : Path of the folder containing originals folder and labels folder to be loaded in dataset.
            Folder names must be as mentioned.
        n_classes (int)  : number of classes in label images.



    """

    def __init__(self, path, n_classes):

        images_path = os.path.join(path, "originals")
        labels_path = os.path.join(path, "labels")
        assert os.path.isdir(
            images_path), "Path does not contian originals folder."
        assert os.path.isdir(
            labels_path), "Path does not contian labels folder."

        img_paths, label_paths = self._read_img_label_paths(
            images_path,
            labels_path)
        self.n_classes = n_classes
        self.size = len(img_paths)

        ds = tf.data.Dataset.from_tensor_slices((img_paths, label_paths))
        ds = ds.map(self._decode_and_resize)
        ds = ds.apply(tf.data.experimental.ignore_errors())

        self.dataset = ds
        print("Dataset build....")

    def _decode_and_resize(self, image_filename, label_filename):
        r""" Reads images. Reads and one hot encodes labels.

        Args:
            image_filename (str) : Path of image file.
            label_filename (str) : Path of label file.

        Returns:
            :class:`tensorflow.Tensor`  : Image tensor.
            :class:`tensorflow.Tensor`  : Label tensor.
        """
        image = tf.io.read_file(image_filename)
        image = tf.io.decode_png(image, channels=3)
        image = tf.image.convert_image_dtype(image, tf.float32)
        resize = (tf.shape(image[:, :, 0])//32)*32
        image = tf.image.resize(image, resize)
        label = tf.io.read_file(label_filename)
        label = tf.io.decode_png(label, channels=3)
        label = tf.image.resize(label, resize)
        label = tf.cast((label[:, :, 0] > 100), tf.int32)
        label = tf.one_hot(label, self.n_classes)
        return image, label

    def __call__(self, batch_size=1, repeat=1):
        r"""

        Args:
            batch_size (int,optional) : Batchsize of :class:`~tf.data.datset`. 
                Default value 1.
            repeat (int, optional)    : Specifies the number of times the dataset will  be iterated in one epoch.
                Default value 1


        Return:
            :class:`tensorflow.data.dataset`  : Dataloader.
        """
        ds = self.dataset
        ds = ds.batch(batch_size).repeat(repeat)
        ds = ds.prefetch(tf.data.experimental.AUTOTUNE)
        return ds

    def __len__(self):
        r"""
        Return:
            int:lenght of dataset
        """
        return self.size

    def _read_img_label_paths(self, images_path, labels_path):
        r"""Reads paths of images and labels.


        Args:
            images_path (str) : Path of the folder of images.
            labels_path (str) : Path of the folder of labels.

        Returns:
            list  : List of image paths.
            list : List of label paths.

        """
        img_path = os.listdir(images_path)
        img_path.sort()
        label_path = os.listdir(labels_path)
        label_path.sort()
        check = [True if i.split('.')[0] == j.split(
            ".")[0] else False for i, j in zip(img_path, label_path)]
        assert tf.reduce_all(
            check), "Originals and Labels does not contain  images with same filenames."
        img_path = [str(os.path.join(images_path, i)) for i in img_path]
        label_path = [str(os.path.join(labels_path, i)) for i in label_path]

        return img_path, label_path


class OCRDataLoader():
    r""" Loads the :class:`~tensorflow.data.Dataset` for :class:`~batukh.tensorflow.ocr.OCR` class.

    Example
    .. code:: python

        >>> from batukh.tensorflow.utils.data.dataloader import OCRDataLoader
        >>> dl=OCRDataLoader("/data/",64)
        >>> for i,j in dl(1,1):
        ...     print(i.shape,j.values)
        ...     break

    Args:
        path (str)        :  Path of  folder containing images folder,labels.txt and table.txt to be loaded in dataset.Name of folders and files should be same as mentioned.
        height (int,optional)      : Specifies the height to which all images will be resized keeping same aspect ratio.
            Default: :math:`64`
    """

    def __init__(self, path, height=64):

        images_path = os.path.join(path, "images")
        labels_path = os.path.join(path, "labels.txt")
        table_path = os.path.join(path, "table.txt")
        assert os.path.isdir(
            images_path), "Path does not contian images folder."
        assert os.path.exists(
            labels_path), "Path does not contian label.txt."
        assert os.path.exists(
            labels_path), "Path does not contian table.txt."
        img_paths, labels = self._read_img_paths_and_labels(
            images_path,
            labels_path)
        self.size = len(img_paths)
        self.path = path
        self.height = height

        with open(table_path) as f:
            self.inv_table = [char.strip() for char in f]
        self.n_classes = len(self.inv_table)
        self.blank_index = self.n_classes - 1

        self.table = tf.lookup.StaticHashTable(tf.lookup.TextFileInitializer(
            table_path, tf.string, tf.lookup.TextFileIndex.WHOLE_LINE,
            tf.int64, tf.lookup.TextFileIndex.LINE_NUMBER), self.blank_index)

        ds = tf.data.Dataset.from_tensor_slices((img_paths, labels))
        ds = ds.map(self._decode_and_resize)
        ds = ds.apply(tf.data.experimental.ignore_errors())

        self.dataset = ds
        print("Dataset build....")

    def _decode_and_resize(self, filename, label):
        r""" Reads image.

        Args:
            filename (str) : Name of file.
            label    (str) : Label of  image.

        Returns:
            :class:`tensorflow.Tensor` : Image Tensor
            :class:`tensorflow.tf.Tensor` :  Label tensor

        """
        image = tf.io.read_file(filename)
        image = tf.io.decode_png(image, channels=1)
        image = 1.0-tf.image.convert_image_dtype(image, tf.float32)
        image = tf.image.resize(
            image, (self.height, tf.shape(image)[-2]), preserve_aspect_ratio=True)
        return image, label

    def _convert_label(self, image, label):
        r""" Maps chars in label to integers  according to table.txt

        Args:
            image ( :class:`tensorflow.tensor`) : Image tensor
            label  (str)      : label

        Returns:
            :class:`tensorflow.Tensor` : Image tensor. 
            :class:`tensorflow.Tensor` : Label sparse tensor.
        """
        chars = tf.strings.unicode_split(label, input_encoding="UTF-8")
        mapped_label = tf.ragged.map_flat_values(self.table.lookup, chars)
        sparse_label = mapped_label.to_sparse()
        label = tf.cast(sparse_label, tf.int32)
        return image, label

    def __len__(self):
        r"""
        Return:
            int:lenght of dataset
        """
        return self.size

    def __call__(self, repeat=1):
        r"""

        Args:
            batch_size (int,optional): specifies the batch size.
                Default: :math:`1`
            repeat (int,optional): Specifes the of time :class:`tensorflow.data.dataset` will be itterated in one epoch.
                Default: :math:`1`
        Returns:
            :class:`tensorflow.data.dataset` : Dataloader.
            """
        ds = self.dataset
        ds = ds.batch(1).map(
            self._convert_label).repeat(repeat)
        ds = ds.prefetch(tf.data.experimental.AUTOTUNE)
        return ds

    def _read_img_paths_and_labels(self, images_path, labels_path):
        r"""reads filenames and respective labels.

        Args:
            imgages_path (str)  : Path of image folder.
            labels_path (str)  : path of label.txt
        Returns:
            list   : List of  images filenames.
            list      : List of labels.
        """
        img_path_ = os.listdir(images_path)
        img_path_.sort()

        label_ = open(labels_path, 'r')
        labels_ = label_.readlines()
        labels_ = [labels_[i].split(":")[1].strip()
                   for i in range(len(labels_))]
        labels = []
        check = [True if 0 <= int(i.split(".")[0]) < len(
            labels) else False for i in img_path_]
        img_path = [os.path.join(images_path, i) for i in img_path_]
        assert tf.reduce_all(
            check), "Image filenames doesnt match with any line number in labels.txt"

        for i in img_path:
            labels.append(labels_[int(i.split(".")[0].split("/")[-1])])
        return img_path, labels
